"""
Default settings for net_promoter_score.

Override these settings in the Django settings if required.

"""

from __future__ import annotations

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest

from .models import UserScore


def default_display_function(request: HttpRequest) -> bool:
    """
    Return default NPS display function.

    The default display function uses the days_since_user_score to determine
    whether to show NPS to the user. If the interval since the last survey is
    more than the NPS_DISPLAY_INTERVAL, *or* they've never been surveyed, then
    return True.

    """
    if not hasattr(request, "user"):
        raise ImproperlyConfigured(
            "Missing middleware: "
            "'django.contrib.auth.middleware.AuthenticationMiddleware'"
        )
    interval = UserScore.objects.days_since_user_score(request.user)
    if interval == -1:
        return True
    if interval > NPS_DISPLAY_INTERVAL:
        return True
    return False


# default display interval is 30 days
NPS_DISPLAY_INTERVAL = getattr(settings, "NPS_DISPLAY_INTERVAL", 30)
# default display function is to show every 30 days, ignoring any user attrs.
NPS_DISPLAY_FUNCTION = getattr(
    settings, "NPS_DISPLAY_FUNCTION", default_display_function
)



from django.contrib import admin

from .models import UserScore


class UserScoreAdmin(admin.ModelAdmin):
    """Admin model for UserScore objects."""

    list_display = ("user", "timestamp", "score", "group", "source")
    list_filter = ("timestamp", "group", "source")
    readonly_fields = ("user", "timestamp", "score", "group", "reason")


admin.site.register(UserScore, UserScoreAdmin)



# -*- coding: utf-8 -*-


from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="UserScore",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("timestamp", models.DateTimeField()),
                (
                    "score",
                    models.IntegerField(
                        default=-1,
                        help_text="0-6=Detractor; 7-8=Neutral; 9-10=Promoter",
                        db_index=True,
                    ),
                ),
                (
                    "reason",
                    models.TextField(help_text="Reason for the score", blank=True),
                ),
                (
                    "group",
                    models.CharField(
                        default="unknown",
                        help_text="Detractor, neutral or promoter.",
                        max_length=10,
                        db_index=True,
                        choices=[
                            ("unknown", "No answer"),
                            ("detractor", "Detractor (0-6)"),
                            ("neutral", "Neutral (7-8)"),
                            ("promoter", "Promoter (9-10)"),
                        ],
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        to=settings.AUTH_USER_MODEL,
                        on_delete=models.CASCADE,
                        related_name="nps_scores",
                    ),
                ),
            ],
        )
    ]



from __future__ import annotations

from typing import Any

from django import forms

from .models import UserScore


class UserScoreForm(forms.ModelForm):
    """Form for capturing NPS score."""

    reason = forms.CharField(max_length=512, required=False)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "user" not in kwargs:
            raise ValueError("Score must have a valid user.")
        self.user = kwargs.pop("user")
        super(UserScoreForm, self).__init__(*args, **kwargs)

    class Meta:
        model = UserScore
        fields = ("score", "reason")
        exclude = ("user",)

    def clean_score(self) -> int:
        score = self.cleaned_data["score"]  # type: int
        if score < -1 or score > 10:
            raise forms.ValidationError("Score must be between 0-10")
        return score

    def save(self, commit: bool = True) -> UserScore:
        """Set the user attr of the score."""
        score = super(UserScoreForm, self).save(commit=False)
        score.user = self.user
        return score.save()



try:
    from django.urls import re_path
except ImportError:
    from django.conf.urls import url as re_path

from . import views

app_name = "net_promoter_score"

urlpatterns = [re_path(r"^score/$", views.post_score, name="post_score")]



from __future__ import annotations

from typing import Any

from django.db.models.signals import post_save
from django.dispatch import Signal, receiver

from .models import UserScore

# custom signal sent when a new score is submitted
# providing_args=["instance"]
new_nps_score = Signal()


@receiver(post_save, sender=UserScore)
def on_new_score(
    sender: type[UserScore], instance: UserScore, created: bool, **kwargs: Any
) -> None:
    # this is a pass-through that filters just new objects
    if not created:
        return
    new_nps_score.send(sender=sender, instance=instance)



from __future__ import annotations

from django.http import HttpRequest

from . import settings


def show_nps(request: HttpRequest) -> bool:
    """
    Return True if the NPS survey should be shown to the request user.

    This is a dynamic function that uses the request object to determine
    whether to display the survey. It defaults to a basic function that
    looks up the last time the user was shown the survey, so that each
    user is only shown it every X days. This should be overridden in the
    Django settings to provide more sophisticated analysis - users with
    different profiles may be surveyed more/less frequently, and you may
    wish to survery based on previous answers.

    """
    return settings.NPS_DISPLAY_FUNCTION(request)



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("net_promoter_score", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="userscore",
            name="source",
            field=models.CharField(
                help_text="Source of user score, used for filtering results, e.g. 'app', 'web', 'email'.",
                max_length=20,
                blank=True,
            ),
        )
    ]



from __future__ import annotations

from typing import Any

from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser
from django.db import models
from django.utils.timezone import now as tz_now


def score_group(score: int) -> str:
    """Return 'detractor', 'neutral' or 'promoter'."""
    if not isinstance(score, int):
        raise ValueError("Score must be an integer value")
    if score < -1:
        raise ValueError("Score cannot be less than -1")
    if score > 10:
        raise ValueError("Score cannot be greater than 10")

    if score == -1:
        return UserScore.GROUP_UNKNOWN
    elif score < 7:
        return UserScore.GROUP_DETRACTOR
    elif score < 9:
        return UserScore.GROUP_NEUTRAL
    else:
        return UserScore.GROUP_PROMOTER


class UserScoreQuerySet(models.query.QuerySet):
    """Custom UserScore queryset - calculates the NPS."""

    def promoters(self) -> UserScoreQuerySet:
        """Filter queryset by promoters."""
        return self.filter(group=UserScore.GROUP_PROMOTER)

    def detractors(self) -> UserScoreQuerySet:
        """Filter queryset by detractors."""
        return self.filter(group=UserScore.GROUP_DETRACTOR)

    def neutrals(self) -> UserScoreQuerySet:
        """Filter queryset by neutrals."""
        return self.filter(group=UserScore.GROUP_NEUTRAL)

    def net_promoter_score(self) -> float:
        """
        Calculate the NPS score from a set of UserScore objects.

        The NPS is the % of promoters - % of detractors (ignoring
        neutrals).

        """
        promoters = self.promoters().count()
        detractors = self.detractors().count()
        neutrals = self.neutrals().count()
        total = promoters + detractors + neutrals
        return (100.0 * (promoters - detractors)) / total

    def most_recent_user_score(self, user: AbstractBaseUser) -> UserScore:
        """Return the most recent user UserScore."""
        return self.filter(user=user).order_by("-id").first()

    def days_since_user_score(self, user: AbstractBaseUser) -> int:
        """
        Return the number of days since the User last submitted a score.

        Returns -1 if the user has never been asked.

        """
        score = self.most_recent_user_score(user)
        return score.elapsed if score else -1


class UserScore(models.Model):
    """Records the NPS score given by a user."""

    GROUP_UNKNOWN = "unknown"
    GROUP_PROMOTER = "promoter"
    GROUP_DETRACTOR = "detractor"
    GROUP_NEUTRAL = "neutral"

    GROUP_CHOICES = (
        (GROUP_UNKNOWN, "No answer"),
        (GROUP_DETRACTOR, "Detractor (0-6)"),
        (GROUP_NEUTRAL, "Neutral (7-8)"),
        (GROUP_PROMOTER, "Promoter (9-10)"),
    )

    class Meta:
        app_label = "net_promoter_score"

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        db_index=True,
        related_name="nps_scores",
    )
    timestamp = models.DateTimeField()
    score = models.IntegerField(
        default=-1,
        blank=False,
        null=False,
        # validators=[MinValueValidator(-1), MaxValueValidator(10)],
        help_text="0-6=Detractor; 7-8=Neutral; 9-10=Promoter",
        db_index=True,
    )
    reason = models.TextField(blank=True, help_text="Reason for the score")
    # this field is calculated from the score, but is denormalised into
    # a separate property to make it easier to report on. It's kept in
    # sync in the save method.
    group = models.CharField(
        max_length=10,
        default=GROUP_UNKNOWN,
        choices=GROUP_CHOICES,
        help_text="Detractor, neutral or promoter.",
        db_index=True,
    )
    source = models.CharField(
        max_length=20,
        blank=True,
        help_text=(
            "Source of user score, used for filtering results, "
            "e.g. 'app', 'web', 'email'."
        ),
    )

    objects = UserScoreQuerySet.as_manager()

    def __str__(self) -> str:
        if self.timestamp:
            return f"{self.group} ({self.score}/10), {self.timestamp.date()}"
        else:
            return f"{self.group} ({self.score}/10), unsaved"

    def __repr__(self) -> str:
        return f"<UserScore: id={self.pk} score={self.score}>"

    def save(self, *args: Any, **kwargs: Any) -> UserScore:
        """Set the timestamp and group attrs."""
        self.timestamp = self.timestamp or tz_now()
        self.group = score_group(self.score)
        super().save(*args, **kwargs)
        return self

    def json(self) -> dict:
        """Return JSON-serializable form of the object."""
        if not self.id:
            raise ValueError("Score must be saved before calling json().")
        return {
            "id": self.id,
            "user": self.user.pk,
            "score": self.score,
            "group": self.group,
        }

    @property
    def is_promoter(self) -> bool:
        return self.group == UserScore.GROUP_PROMOTER

    @property
    def is_detractor(self) -> bool:
        return self.group == UserScore.GROUP_DETRACTOR

    @property
    def is_neutral(self) -> bool:
        return self.group == UserScore.GROUP_NEUTRAL

    @property
    def is_unknown(self) -> bool:
        return self.group == UserScore.GROUP_UNKNOWN

    @property
    def elapsed(self) -> int:
        """Return number of days since the score was submitted."""
        return (tz_now().date() - self.timestamp.date()).days



"""
NPS middleware used to set request attrs and cookies.

When determining whether to display the NPS survey to user, we cache
the output (True|False) so that we don't have to do a database lookup
on each request. This value is added to the user session (so a max
of one lookup per session.

"""

from __future__ import annotations

from typing import Callable

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponse

from .utils import show_nps


class NPSMiddleware:
    def __init__(self, get_response: Callable) -> None:
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if not hasattr(request, "user"):
            raise ImproperlyConfigured(
                "Missing middleware: "
                "'django.contrib.auth.middleware.AuthenticationMiddleware'"
            )
        if request.user.is_authenticated:
            request.show_nps = show_nps(request)
        else:
            request.show_nps = False

        return self.get_response(request)
